import mlx.core as mx
import mlx.nn as nn
def length_masked_ce_loss(model, inputs, targets, lengths):
  # Run model on inputs
  logits = model(inputs).astype(mx.float32)
  
  # Mask padding tokens
  length_mask = mx.arange(inputs.shape[1])[None, :] < lengths[:, None]

  # Calculate the loss
  ce = nn.losses.cross_entropy(logits, targets) * length_mask
  loss = ce.sum() / length_mask.sum()
#  print(f"|    {inputs=}\n| ==>{logits=}\n| ~^~{ce=}\n| == {loss=}")
  return loss

#Naive intermediate layer loss, where we replace the targets with gradients and just multiply the output by the gradients to derive the loss. This is naive and may warrant some further iteration, but will do the job for now
def back_gradient_loss(model, inputs, gradients, lengths):
  out = model(inputs).astype(mx.float32)
  grad = gradients.astype(mx.float32)

  # Mask padding tokens
  length_mask = mx.repeat(mx.arange(inputs.shape[1])[None, :] < lengths[:, None], out.shape[-1]).reshape(out.shape)

  masked_sum = (out * length_mask).sum(axis=1)
  gradient_lens = mx.abs(grad * masked_sum)
  loss = gradient_lens.sum() / length_mask.sum()
#  print(f"|    {inputs=}\n"
#      + f"| ==>{out=}\n"
#      + f"| ~^~{masked_sum=}\n"
#      + f"| <~>{gradient_lens=}\n"
#      + f"| == {loss=}")
  return loss

loss_fns = {
  "back_gradient": back_gradient_loss,
  "length_masked_ce": length_masked_ce_loss,
}



from net.inference.inference_engine import InferenceEngine
from net.inference.shard import Shard
from net.inference.tinygrad.inference import TinygradDynamicShardInferenceEngine
import asyncio
import numpy as np


# An inference engine should work the same for any number of Shards, as long as the Shards are continuous.
async def test_inference_engine(inference_engine_1: InferenceEngine, inference_engine_2: InferenceEngine, model_id: str):
  from net.inference.tinygrad.inference import Tokenizer
  from pathlib import Path

  _tokenizer = Tokenizer(str(Path(model_id)/"tokenizer.model"))

  prompt = "In a single word only, what is the last name of the president of the United States? "
  resp_full = await inference_engine_1.infer_prompt("A", shard=Shard(model_id=model_id, start_layer=0, end_layer=31, n_layers=32), prompt=prompt)
  token_full = await inference_engine_1.sample(resp_full)

  next_resp_full, _ = await inference_engine_1.infer_tensor(
    "A",
    shard=Shard(model_id=model_id, start_layer=0, end_layer=31, n_layers=32),
    input_data=token_full,
  )

  resp1, _ = await inference_engine_1.infer_prompt("B", shard=Shard(model_id=model_id, start_layer=0, end_layer=30, n_layers=32), prompt=prompt)
  resp2, _ = await inference_engine_2.infer_tensor(
    "B",
    shard=Shard(model_id=model_id, start_layer=31, end_layer=31, n_layers=32),
    input_data=resp1,
  )
  token2 = await inference_engine_2.sample(resp2)
  resp3, _ = await inference_engine_1.infer_tensor(
    "B",
    shard=Shard(model_id=model_id, start_layer=0, end_layer=30, n_layers=32),
    input_data=token2,
  )
  resp4, _ = await inference_engine_2.infer_tensor(
    "B",
    shard=Shard(model_id=model_id, start_layer=31, end_layer=31, n_layers=32),
    input_data=resp3,
  )

  print(f"{resp2=}")
  print(f"full: {_tokenizer.decode(resp_full)}")
  print(f"next full: {_tokenizer.decode(next_resp_full)}")
  print(f"resp2: {_tokenizer.decode(resp2)}")
  print(f"{resp4=}")
  print(f"resp4: {_tokenizer.decode(resp4)}")

  assert np.array_equal(resp_full, resp2)
  assert np.array_equal(next_resp_full, resp4)


asyncio.run(test_inference_engine(
  TinygradDynamicShardInferenceEngine(),
  TinygradDynamicShardInferenceEngine(),
  "llama3-8b-sfr",
))



import pytest
import numpy as np
from net.inference.dummy_inference_engine import DummyInferenceEngine
from net.inference.shard import Shard


@pytest.mark.asyncio
async def test_dummy_inference_specific():
  engine = DummyInferenceEngine()
  test_shard = Shard(model_id="test_model", start_layer=0, end_layer=1, n_layers=1)
  test_prompt = "This is a test prompt"

  result, _ = await engine.infer_prompt("test_request", test_shard, test_prompt)

  print(f"Inference result shape: {result.shape}")

  assert result.shape[0] == 1, "Result should be a 2D array with first dimension 1"


@pytest.mark.asyncio
async def test_dummy_inference_engine():
  # Initialize the DummyInferenceEngine
  engine = DummyInferenceEngine()

  # Create a test shard
  shard = Shard(model_id="test_model", start_layer=0, end_layer=1, n_layers=1)

  # Test infer_prompt
  output, _ = await engine.infer_prompt("test_id", shard, "Test prompt")

  assert isinstance(output, np.ndarray), "Output should be a numpy array"
  assert output.ndim == 2, "Output should be 2-dimensional"

  # Test infer_tensor
  input_tensor = np.array([[1, 2, 3]])
  output, _ = await engine.infer_tensor("test_id", shard, input_tensor)

  assert isinstance(output, np.ndarray), "Output should be a numpy array"
  assert output.ndim == 2, "Output should be 2-dimensional"

  print("All tests passed!")


if __name__ == "__main__":
  import asyncio
  asyncio.run(test_dummy_inference_engine())
  asyncio.run(test_dummy_inference_specific())
